from __future__ import absolute_import

import os
import sys

from ._ros_generator import (
    MsgDependencyNotFound,
    genrosmsg_py,
    genrossrv_py,
)

from ._rosdef_loader import ROSMsgLoader, ROSSrvLoader, ros_import_search_path

from ._ros_directory_finder import get_supported_ros_loaders, ROSDirectoryFinder, ROSPathFinder

from ._utils import _verbose_message

ros_path_hook = ROSDirectoryFinder.path_hook(*get_supported_ros_loaders())


def activate():
    """Install the path-based import components."""
    if sys.version_info < (3, 4):
        # We should plug filefinder first to avoid plugging ROSDirectoryFinder, when it is not a ROS thing...
        import filefinder2
        filefinder2.activate()
        PathFinder = filefinder2.NamespaceMetaFinder2

        if ros_path_hook not in sys.path_hooks:
            # We need to be before FileFinder to be able to find our '.msg' and '.srv' files without making a namespace package
            # Note this must be early in the path_hook list, since we change the logic
            # and a namespace package becomes a ros importable package.
            sys.path_hooks.insert(sys.path_hooks.index(filefinder2.path_hook), ros_path_hook)
    else:
        from importlib.machinery import PathFinder

        if ros_path_hook not in sys.path_hooks:
            # Note : On py 3.5 the import system doesnt use hooks after FileFinder since it assumes
            # any directory not containing __init__.py is a namespace package
            sys.path_hooks.insert(1, ros_path_hook)

    if ROSPathFinder not in sys.meta_path:
        # adding metahook, before the usual pathfinder, to avoid interferences with python namespace mechanism...
        sys.meta_path.insert(sys.meta_path.index(PathFinder), ROSPathFinder)

    # Resetting sys.path_importer_cache
    # to support the case where we have an implicit (msg/srv) package inside an already loaded package,
    # since we need to replace the default importer.
    sys.path_importer_cache.clear()


def deactivate():
    # CAREFUL : Even though we remove the path from sys.path,
    # initialized finders will remain in sys.path_importer_cache

    # removing metahook
    sys.meta_path.remove(ROSPathFinder)
    sys.path_hooks.remove(ros_path_hook)

    # Resetting sys.path_importer_cache to get rid of previous importers
    sys.path_importer_cache.clear()


__all__ = [
    'MsgDependencyNotFound',
    'generate_msgsrv_nspkg',
    'ROSMsgLoader',
    'ROSSrvLoader',
    'genmsg_py',
    'gensrv_py',
    'activate_hook_for',
    'deactivate_hook_for',
]



from __future__ import absolute_import, division, print_function

import os
import sys
import tempfile
import traceback
import importlib
import site
import time

import collections
import pkg_resources

"""
Module that can be used standalone, or as part of the rosimport package
It provides a set of functions to generate your ros messages, even when ROS is not installed on your system.

This module allows to generate all rosdef message/service python code for one ros package, including subpackages.

This module does NOT provide a solution for one package to depend on another.
That problem will be handled by leveraging the python import system.

Note : Generated modules/packages can only be imported once. So it is important to provide an API that : 
- makes it easy to generate the whole module/package at once, since this is our priority
- makes it easy to optionally import the whole generated module/package
- still allows to generate only one module / a part of the whole package, caveats apply / warning added.
"""

try:
    # Using genpy and genmsg directly if ROS has been setup (while using from ROS pkg)
    import genmsg as genmsg
    import genmsg.command_line as genmsg_command_line
    import genpy.generator as genpy_generator
    import genpy.generate_initpy as genpy_generate_initpy

except ImportError:

    # Otherwise we refer to our submodules here (setup.py usecase, or running from tox without site-packages)

    import site
    ros_site_dir = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'ros-packages')
    print("Adding site directory {ros_site_dir} to access genpy and genmsg.".format(**locals()))
    site.addsitedir(ros_site_dir)

    import genmsg as genmsg
    import genmsg.command_line as genmsg_command_line
    import genpy.generator as genpy_generator
    import genpy.generate_initpy as genpy_generate_initpy

    # Note we do not want to use pyros_setup here.
    # We do not want to do a full ROS setup, only import specific packages.
    # If needed it should have been done before (loading a parent package).
    # this handle the case where we want to be independent of any underlying ROS system.

# TMP bwcompat
try:
    import genmsg.MSG_DIR as genmsg_MSG_DIR
    import genmsg.SRV_DIR as genmsg_SRV_DIR
except ImportError:  # bwcompat
    genmsg_MSG_DIR = 'msg'
    genmsg_SRV_DIR = 'srv'


class MsgGenerationFailed(Exception):
    pass


class MsgDependencyNotFound(Exception):
    pass


class PkgAlreadyExists(Exception):
    pass


def _generator_factory(generator, directory_name, file_extension):

    def _generator_py_pkg(files, package, outdir, search_path=None, initpy=True):
        """
        Generates python code from ROS definition files
        :param files: the list of ros definition files to generate from
        :param package: the ROS package for which we are generating messages / services
        :param outdir: the directory where to output the code generated for this package.
        :param search_path: a dict where keys are ROS package names, and value is a list of path to directory containing '.msg' files
        :param initpy: Whether or not generate the __init__.py for the package
        :return:
        """

        # Note only msg files can be used as dependencies.
        # setting search path here ensures it is stored in ros_search_path
        search_path = {} if search_path is None else search_path

        # checking if we have files with unknown extension to except early
        for f in files:
            if not f.endswith(file_extension):
                print("WARNING: {f} doesnt have the proper {file_extension} extension. It has been Ignored.".format(**locals()), file=sys.stderr)

        filtered_files = [f for f in files if f.endswith(file_extension)]
        genset = set()

        if filtered_files:

            if not os.path.exists(outdir):
                # This script can be run multiple times in parallel. We
                # don't mind if the makedirs call fails because somebody
                # else snuck in and created the directory before us.
                try:
                    os.makedirs(outdir)
                except OSError as e:
                    if not os.path.exists(outdir):
                        raise
            try:
                retcode = generator.generate_messages(
                    package=package,
                    package_files=filtered_files,
                    outdir=outdir,
                    search_path=search_path
                )
                assert retcode == 0

                # because the OS interface might not be synchronous....
                while not os.path.exists(outdir):
                    time.sleep(.1)

            except genmsg.InvalidMsgSpec as e:
                print("ERROR: ", e, file=sys.stderr)
                raise
            except genmsg.MsgGenerationException as e:
                print("ERROR: ", e, file=sys.stderr)
                raise
            except Exception as e:
                traceback.print_exc()
                print("ERROR: ", e)
                raise

            # optionally we can generate __init__.py
            if initpy:
                init_path = os.path.join(outdir, '__init__.py')
                # Note if it already exists, we overwrite it. This should accumulate generated modules.
                genpy_generate_initpy.write_modules(outdir)
                genset.add(init_path)
            else:  # we list all files, only if init.py was not created (and user has to import one by one)
                for f in files:
                    f, _ = os.path.splitext(f)  # removing extension
                    os.path.relpath(outdir)
                    genset.add(os.path.join(outdir, '_' + os.path.basename(f) + '.py'))

        return genset

    return _generator_py_pkg

# TODO : get extensions and dir from genmsg
_genmsgpkg_py = _generator_factory(genpy_generator.MsgGenerator(), 'msg', '.msg')
_gensrvpkg_py = _generator_factory(genpy_generator.SrvGenerator(), 'srv', '.srv')


def genrosmsg_py(rosdef_files, package, sitedir, search_path=None):
    """
    Generates message/services modules for a package, in that package directory,
    in a subpackage called 'msg'/'srv', following ROS conventions
    :param rosdef_files: the .msg/.srv files to use as input for generating the python message classes
    :param package: the package for which we want to generate these messages.
    This is a string separated with '.', the head element being the ros package,
    the tail being the optional python subpackage where the code will be generated.
    :param sitedir: the site directory where to put the generated package.
    :param search_path: optionally a mapping of the form {package: [list of paths]} , in order to retrieve message dependencies
    Note that all dependencies must have been previously generated, or passed in the rosdef_files list,
    otherwise generation will fail if not included in search_path, or import will fail afterwards...
    :return: the list of files generated
    """

    # Computing outdir_pkg from sitedir and package
    rospackage = package.partition('.')[0]

    outdir = sitedir
    for pkg in package.split('.'):
        # removing the trailing 'msg' if present, to enforce it if not
        if not (package.endswith(pkg) and pkg == 'msg'):
            outdir = os.path.join(outdir, pkg)

    search_path = {} if search_path is None else search_path

    # modifying the search path for later retrieval of msg as dependencies,
    # if the caller keep it around...
    search_path.setdefault(rospackage, {os.path.dirname(m) for m in rosdef_files})
    # Need to be done before generation to avoid useless harmful recursive self imports

    # Using tuple unpacking to get the unique element of the set (since we ask for init.py)
    (generated_pkg,) = _genmsgpkg_py(
        files=[f for f in rosdef_files if f.endswith('.msg')],
        package=rospackage,
        outdir=os.path.join(outdir, 'msg'),
        search_path=search_path,
        initpy=True
    )

    return sitedir, generated_pkg


def genrossrv_py(rosdef_files, package, sitedir, search_path=None):
    """
    Generates message/services modules for a package, in that package directory,
    in a subpackage called 'msg'/'srv', following ROS conventions
    :param rosdef_files: the .msg/.srv files to use as input for generating the python message classes
    :param package: the package for which we want to generate these messages.
    This is a string separated with '.', the head element being the ros package,
    the tail being the optional python subpackage where the code will be generated.
    :param sitedir: the site directory where to put the generated package.
    :param search_path: optionally a mapping of the form {package: [list of paths]} , in order to retrieve message dependencies
    Note that all dependencies must have been previously generated, or passed in the rosdef_files list,
    otherwise generation will fail if not included in search_path, or import will fail afterwards...
    :return: the list of files generated
    """

    # Computing outdir_pkg from sitedir and package
    rospackage = package.partition('.')[0]

    outdir = sitedir
    for pkg in package.split('.'):
        # removing the trailing 'srv' if present, to enforce it if not
        if not (package.endswith(pkg) and pkg == 'srv'):
            outdir = os.path.join(outdir, pkg)

    search_path = search_path or {}

    # Here the include path has already been added to ros_search_path,
    # and can be used to discover msg as dependencies
    # Using tuple unpacking to get the unique element of the set (since we ask for init.py)
    (generated_pkg, ) = _gensrvpkg_py(
        files=[f for f in rosdef_files if f.endswith('.srv')],
        package=rospackage,
        outdir=os.path.join(outdir, 'srv'),
        search_path=search_path,
        initpy=True
    )

    return sitedir, generated_pkg


# def generate_rosdefs_py(files, package, sitedir=None):
#     """
#     Generates ros messages python modules for a set of rosdefs files.
#     :param files: the rosdef files to generate python code from
#     :param package: the package these rosdefs belong to.
#     This is a string separated with '.', the head element being the ros package,
#     the tail being the optional python subpackage where the code will be generated.
#     :param sitedir: the output dir where to put the generated code for the (base / ros) package
#     :return:
#     """
#     # TODO : since we return a full package, we should probably pass a dir, not the files one by one...
#     # by default we generate for this package (does it make sense ?)
#
#     sitedir = sitedir or tempfile.mkdtemp('rosimport_generated_site')
#
#     # If our sitedir is not inside an already registered path
#     parent_in_sys_path = False
#     sitedir_splitted = sitedir.split(os.sep)
#     for i, _ in enumerate(sitedir_splitted):
#         if os.path.join('/' if sitedir_splitted[0] == '' else '', *sitedir_splitted[:len(sitedir_splitted)-i]) in sys.path:
#             parent_in_sys_path = True
#             break
#     if not parent_in_sys_path:
#         # We need to add the parent of our generated directory, for any messages dependency to resolve properly
#         # and be able to import from it as usual, even while generating python code...
#         site.addsitedir(sitedir)  # we add our output dir as a site
#         # This is required to be able to access generated python code
#
#     outdir, gen_files = genros_py(
#         rosdef_files=files,
#         package=package,
#         sitedir=sitedir,
#     )
#
#     # computing module names that will be importable after outdir has been added as sitedir
#     gen_msgs = None
#     gen_srvs = None
#     # Not ideal, but it will do until we implement a custom importer
#     for f in gen_files:
#         f = f[:-len('.py')] if f.endswith('.py') else f
#         f = f[:-len(os.sep + '__init__')] if f.endswith(os.sep + '__init__') else f
#
#         if f.endswith('msg'):
#             gen_msgs = package + '.msg'
#
#         if f.endswith('srv'):
#             gen_srvs = package + '.srv'
#
#     # we need to get one level up to get the sitedir (containing the generated namespace package)
#     return sitedir, gen_msgs, gen_srvs




from __future__ import absolute_import, division, print_function

import contextlib
import importlib
import site
import tempfile

import shutil


from rosimport import genrosmsg_py, genrossrv_py

"""
A module to setup custom importer for .msg and .srv files
Upon import, it will first find the .msg file, then generate the python module for it, then load it.

TODO...
"""

# We need to be extra careful with python versions
# Ref : https://docs.python.org/dev/library/importlib.html#importlib.import_module

# Ref : http://stackoverflow.com/questions/67631/how-to-import-a-module-given-the-full-path
# Note : Couldn't find a way to make imp.load_source deal with packages or relative imports (necessary for our generated message classes)
import os
import sys
import logging


# Class to allow dynamic search of packages
class RosSearchPath(dict):
    """
    Class to allow dynamic search of packages.
    This is where we hook up into python import mechanism in order to generate and discover
    packages and messages we are depending on.

    But it should not be used during the generation of multiple messages in only one package,
    as this is too tricky to get right, and too easy to break by mistake.
    """
    def __init__(self, **ros_package_paths):
        # we use the default ROS_PACKAGE_PATH if setup in environment
        package_paths = {}  # TODO : os.environ.get('ROS_PACKAGE_PATH', {}) OR NOT ?
        # we add any extra path
        package_paths.update(ros_package_paths)
        super(RosSearchPath, self).__init__(package_paths)

    def try_import(self, item):
        try:
            # we need to import the .msg submodule (only one usable as dependency)
            mod = importlib.import_module(item + '.msg')
            # import succeeded : we should get the namespace path
            # and add it to the list of paths to avoid going through this all over again...
            for p in mod.__path__:
                # Note we want dependencies here. dependencies are ALWAYS '.msg' files in 'msg' directory.
                msg_path = os.path.join(p)
                # We add a path only if we can find the 'msg' directory
                self[item] = self.get(item, set() | ({msg_path} if os.path.exists(msg_path) else set()))
            return mod
        except ImportError:
            # import failed
            return None

    def __contains__(self, item):
        """ True if D has a key k, else False. """
        has = super(RosSearchPath, self).__contains__(item)
        if not has:  # attempt importing. solving ROS path setup problem with python import paths setup.
            self.try_import(item)
            # Note : if ROS is setup, rospkg.RosPack can find packages
        # try again (might work now)
        return super(RosSearchPath, self).__contains__(item)

    def __getitem__(self, item):
        """ x.__getitem__(y) <==> x[y] """
        got = super(RosSearchPath, self).get(item)
        if got is None:
            # attempt discovery by relying on python core import feature.
            self.try_import(item)
            # Note : if ROS is setup, rospkg.RosPack can find packages
        return super(RosSearchPath, self).get(item)

# singleton instance, to keep used ros package paths in cache
ros_import_search_path = RosSearchPath()


def RosLoader(rosdef_extension):
    """
    Function generating ROS loaders.
    This is used to keep .msg and .srv loaders very similar
    """
    if rosdef_extension == '.msg':
        loader_origin_subdir = 'msg'
        loader_file_extension = rosdef_extension
        loader_generated_subdir = 'msg'
        loader_generator = genrosmsg_py
    elif rosdef_extension == '.srv':
        loader_origin_subdir = 'srv'
        loader_file_extension = rosdef_extension
        loader_generated_subdir = 'srv'
        loader_generator = genrossrv_py
    else:
        raise RuntimeError("RosLoader for a format {0} other than .msg or .srv is not supported".format(rosdef_extension))

    if (2, 7) <= sys.version_info < (3, 4):  # valid until which py3 version ?
        import filefinder2
        FileLoader = filefinder2.SourceFileLoader2
    elif sys.version_info >= (3, 4):  # we do not support 3.2 and 3.3 (unsupported but might work ?)
        import importlib.machinery
        FileLoader = importlib.machinery.SourceFileLoader
    else:
        raise ImportError("ros_loader : Unsupported python version")

    class ROSDefLoader(FileLoader):
        """
        Python Loader for Rosdef files.
        Note : We support ROS layout :
        - msg/myMsg.msg
        - srv/mySrv.srv
        - my_pkg/__init__.py  # doesnt really matters ( we rely on PEP 420 )
        OR inside the python code:
        - my_pkg/__init__.py  # doesnt really matters ( we rely on PEP 420 )
        - my_pkg/msg/myMsg.msg
        - my_pkg/srv/mySrv.srv

        BUT the following is also importable relatively,
        which is especially useful for tests or intra-package ROS communication,
        although it cannot be used as another package dependency (due to ROS limitations)

        - my_pkg/__init__.py  # doesnt really matters ( we rely on PEP 420 )
        - my_pkg/subpkg/__init__.py # doesnt really matters ( we rely on PEP 420 )
        - my_pkg/subpkg/msg/myMsg.msg
        - my_pkg/subpkg/srv/mySrv.srv

        In that case myMsg.py will also be generated under mypkg.msg,
        but can be imported relatively from my_pkg/subpkg/module.py with "from .msg import mypkg"
        """

        rosimport_tempdir = os.path.join(tempfile.gettempdir(), 'rosimport')

        def __init__(self, fullname, path):

            self.logger = logging.getLogger(__name__)
            # to normalize input
            path = os.path.normpath(path)

            # Doing this in each loader, in case we are running from different processes,
            # avoiding to reload from same file (especially useful for boxed tests).
            # But deterministic path to avoid regenerating from the same interpreter
            rosimport_path = os.path.join(self.rosimport_tempdir, str(os.getpid()))
            if not os.path.exists(rosimport_path):
                os.makedirs(rosimport_path)

            rospackage = fullname.partition('.')[0]

            if os.path.isdir(path):
                # if we get a package name ending with msg or srv and a non empty directory
                if (
                            fullname.endswith(loader_origin_subdir) and
                            any([f.endswith(loader_file_extension) for f in os.listdir(path)])
                ):

                    # TODO : dynamic in memory generation (we do not need the file ultimately...)
                    outdir, gen_rosdef_pkgpath = loader_generator(
                        # generate message's python code at once, for this package level.
                        rosdef_files=[os.path.join(path, f) for f in os.listdir(path)],
                        package=fullname,
                        sitedir=rosimport_path,
                        search_path=ros_import_search_path,
                    )
                    # TODO : handle thrown exception (cleaner than hacking the search path dict...)
                    # try:
                    #     generator.generate_messages(package, rosfiles, outdir, search_path)
                    # except genmsg.MsgNotFound as mnf:
                    #     try:
                    #         mod = importlib.import_module(mnf.package)
                    #         # import succeeded : we should get the namespace path that has '/msg'
                    #         # and add it to the list of paths to avoid going through this all over again...
                    #         for p in mod.__path__:
                    #             # Note we want dependencies here. dependencies are ALWAYS '.msg' files in 'msg' directory.
                    #             msg_path = os.path.join(p, genmsg_MSG_DIR)
                    #             # We add a path only if we can find the 'msg' directory
                    #             search_path[mnf.package] = search_path[mnf.package] + ([msg_path] if os.path.exists(msg_path) else [])
                    #         # Try generation again
                    #         generator.generate_messages(package, rosfiles, outdir, search_path)
                    #     except ImportError:
                    #         # import failed
                    #         return None

                    if not os.path.exists(gen_rosdef_pkgpath):
                        raise ImportError("{0} file not found".format(gen_rosdef_pkgpath))

                    # relying on usual source file loader since we have generated normal python code
                    super(ROSDefLoader, self).__init__(fullname, gen_rosdef_pkgpath)

        def get_gen_path(self):
            """Returning the generated path matching the import"""
            return self.path  # TODO : maybe useless ?
            # return os.path.join(self.outdir_pkg, loader_generated_subdir)

        def __repr__(self):
            return "ROSDefLoader/{0}({1}, {2})".format(loader_file_extension, self.name, self.path)

        @staticmethod
        def get_file_extension():
            return loader_file_extension

        @staticmethod
        def get_origin_subdir():
            return loader_origin_subdir

        @staticmethod
        def get_generated_subdir():
            return loader_generated_subdir

    return ROSDefLoader

ROSMsgLoader = RosLoader(rosdef_extension='.msg')
ROSSrvLoader = RosLoader(rosdef_extension='.srv')



from __future__ import absolute_import, division, print_function

import contextlib
import importlib
import site

from rosimport import rosmsg_generator

"""
A module to setup custom importer for .msg and .srv files
Upon import, it will first find the .msg file, then generate the python module for it, then load it.

TODO...
"""

# We need to be extra careful with python versions
# Ref : https://docs.python.org/dev/library/importlib.html#importlib.import_module

# Ref : http://stackoverflow.com/questions/67631/how-to-import-a-module-given-the-full-path
# Note : Couldn't find a way to make imp.load_source deal with packages or relative imports (necessary for our generated message classes)
import os
import sys

# This will take the ROS distro version if ros has been setup
import genpy.generator
import genpy.generate_initpy

import logging





if sys.version_info >= (3, 4):

    import importlib.abc
    import importlib.machinery

    class ROSDirectoryMetaFinder(importlib.machinery.PathFinder):

        @classmethod
        def find_module(cls, fullname, path=None):  # from importlib.PathFinder
            pass

    class ROSDistroMetaFinder(importlib.machinery.PathFinder):
        def __init__(self, *workspaces):
            """
            :param workspaces: can be a devel or install workspace (including a distro install directory), but also a directory containing packages (like a source workspace)
            These should all work, without catkin build necessary.
            """
            super(ROSDistroMetaFinder, self).__init__()
            # TODO : prevent that when we are in virtualenv and we have not allowed system site packages

            self.src_rospkgs = []

            broken_workspaces = []
            for w in workspaces:
                # Adding the share path (where we can find all packages and their messages)
                share_path = os.path.join(w, 'share')
                if os.path.exists(share_path):
                    self.share_path = share_path
                else:  # this is not a workspace, maybe we are expected to get the package directly from source ?
                    found_one = False
                    for root, dirs, files in os.walk(w, topdown=False):
                        if 'package.xml' in files:  # we have found a ros package
                            self.src_rospkgs.append(root)
                            found_one = True
                    if not found_one:
                        broken_workspaces.append(w)

                    raise ImportError

                python_libpath = os.path.join(w, 'lib', 'python' + sys.version_info.major + '.' + sys.version_info.minor)
                if os.path.exists(python_libpath):
                    # adding python site directories for the ROS distro (to find python modules as usual with ROS)
                    if os.path.exists(os.path.join(python_libpath, 'dist-packages')):
                        site.addsitedir(os.path.join(python_libpath, 'dist-packages'))
                    if os.path.exists(os.path.join(python_libpath, 'site-packages')):
                        site.addsitedir(os.path.join(python_libpath, 'site-packages'))
                else:  # this is not a workspace, maybe we are expected to get the package directly from source ?

                    for root, dirs, files in os.walk(w, topdown=False):
                        if 'package.xml' in files:  # we have found a ros package
                            self.src_rospkgs.append(root)

                    raise ImportError




        def find_spec(self, fullname, path, target=None):
            """
            :param fullname: the name of the package we are trying to import
            :param path: path we we expect to find it (can be None)
            :param target: what we plan to do with it
            :return:
            """
            # TODO: read PEP 420 :)
            last_mile = fullname.split('.')[-1]

            path = path or sys.path

            # TODO: we should handle different structure of ROS packages (installed or not)
            #if (os.path.exists)

            # TODO : similar to pyros-setup ?
            # + setup rosmsg importers ?
            if (not os.path.exists(os.path.join(last_mile, 'package.xml')) or
            False):
                raise ImportError

            for p in path:
                fullpath = os.path.join(p, last_mile)
                init_fullpath = os.path.join(fullpath, '__init__.ay')
                module_fullpath = fullpath + '.ay'

                if os.path.isdir(fullpath) and os.path.exists(init_fullpath):
                    return importlib.machinery.ModuleSpec(
                        fullname,
                        loader,
                        origin=init_fullpath
                    )

                else:
                    if os.path.exists(module_fullpath):
                        return importlib.machinery.ModuleSpec(
                            fullname,
                            loader,
                            origin=module_fullpath
                        )

            return None

else:
    passb # TODO


# Useless ?
#_ros_finder_instance_obsolete_python = ROSImportFinder

ros_distro_finder = None


def activate(rosdistro_path=None, *workspaces):
    global ros_distro_finder
    if rosdistro_path is None:  # autodetect most recent installed distro
        if os.path.exists('/opt/ros/lunar'):
            rosdistro_path = '/opt/ros/lunar'
        elif os.path.exists('/opt/ros/kinetic'):
            rosdistro_path = '/opt/ros/kinetic'
        elif os.path.exists('/opt/ros/jade'):
            rosdistro_path = '/opt/ros/jade'
        elif os.path.exists('/opt/ros/indigo'):
            rosdistro_path = '/opt/ros/indigo'
        else:
            raise ImportError(
                "No ROS distro detected on this system. Please specify the path when calling ROSImportMetaFinder()")

    ros_distro_finder = ROSImportMetaFinder(rosdistro_path, *workspaces)
    sys.meta_path.append(ros_distro_finder)

    #if sys.version_info >= (2, 7, 12):  # TODO : which exact version matters ?

    # We need to be before FileFinder to be able to find our (non .py[c]) files
    # inside, maybe already imported, python packages...
    sys.path_hooks.insert(1, ROSImportFinder)

    # else:  # older (trusty) version
    #     sys.path_hooks.append(_ros_finder_instance_obsolete_python)

    for hook in sys.path_hooks:
        print('Path hook: {}'.format(hook))

    # TODO : mix that with ROS PYTHONPATH shenanigans... to enable the finder only for 'ROS aware' paths
    if paths:
        sys.path.append(*paths)


def deactivate(*paths):
    """ CAREFUL : even if we remove our path_hooks, the created finder are still cached in sys.path_importer_cache."""
    #if sys.version_info >= (2, 7, 12):  # TODO : which exact version matters ?
    sys.path_hooks.remove(ROSImportFinder)
    # else:  # older (trusty) version
    #     sys.path_hooks.remove(_ros_finder_instance_obsolete_python)
    if paths:
        sys.path.remove(*paths)

    sys.meta_path.remove(ros_distro_finder)


@contextlib.contextmanager
def ROSImportContext(*paths):
    activate(*paths)
    yield
    deactivate(*paths)


# TODO : a meta finder could find a full ROS distro...



from __future__ import absolute_import, division, print_function

import contextlib
import importlib
import site
from collections import OrderedDict

from rosimport import ROSMsgLoader, ROSSrvLoader

"""
A module to setup custom importer for .msg and .srv files
Upon import, it will first find the .msg file, then generate the python module for it, then load it.

TODO...
"""

# We need to be extra careful with python versions
# Ref : https://docs.python.org/dev/library/importlib.html#importlib.import_module

# Ref : http://stackoverflow.com/questions/67631/how-to-import-a-module-given-the-full-path
# Note : Couldn't find a way to make imp.load_source deal with packages or relative imports (necessary for our generated message classes)
import os
import sys

import logging

from ._utils import _ImportError, _verbose_message




if (2, 7) <= sys.version_info < (3, 4):  # valid until which py3 version ?
    import filefinder2
    PathFinder = filefinder2.PathFinder2
    FileFinder = filefinder2.FileFinder2
elif sys.version_info >= (3, 4):  # we do not support 3.2 and 3.3 (unsupported but might work ?)
    import importlib.machinery
    PathFinder = importlib.machinery.PathFinder
    FileFinder = importlib.machinery.FileFinder
else:
    raise ImportError("ros_loader : Unsupported python version")


class ROSPathFinder(PathFinder):
    """
    MetaFinder to handle Finding ROS package directories
    This is needed because python3 namespace package machinary can miss directories that are
    ROS package root directories, containing msg and srv subdirs.
    """

    def __init__(self, *modules):
        """Initialisation of the finder depending on namespaces.
        The issue here is that we do not have enough information at this stage
        to determine which finder will be most appropriate for this path.

        Only after find_module might we known which one should be chosen."""
        self.module_names = modules

    @classmethod
    def find_spec(cls, fullname, path, target=None):
        loader = cls.find_module(fullname, path)
        spec = importlib.util.spec_from_file_location(
            # will extract all useful information from loader
            fullname,
            loader=loader,
        )
        return spec

    @classmethod
    def find_module(cls, fullname, path=None):  # from importlib.PathFinder
        """Try to find the module on sys.path or 'path'
        The search is based on sys.path_hooks and sys.path_importer_cache.
        """
        if path is None:
            path = sys.path
        loader = None
        for entry in path:
            if not isinstance(entry, (str, bytes)):
                continue
            # first we check if the root import is doable with ROS recursively
            rospkg = fullname.partition('.')[0]
            dirlist = entry.split(os.sep)
            if rospkg in dirlist:
                dirlist[:] = dirlist[:dirlist.index(rospkg)+1]

                rootentry = os.path.join(os.sep if dirlist[0] == '' else '', *dirlist)
                rosentry = os.path.dirname(rootentry)

                # we need to keep order here to maintain strict dependency and generation order
                entrymap = []
                # we are trying to find root messages if we dont have a root package, or if the directory is different
                if rospkg + '.msg' != fullname or rosentry != entry:
                    entrymap.append((rospkg + '.msg', rosentry))
                if rospkg + '.msg' != fullname or rootentry != entry:
                    entrymap.append((rospkg + '.msg', rootentry))
                # This should always be last
                entrymap.append((fullname, entry))

                loader = None
                for n, e in entrymap:
                    finder = cls._path_importer_cache(e)
                    if finder is not None:
                        # creatign the loader will generate the python code and make sure it is importable.
                        l = finder.find_module(n)
                        if l and n == fullname:  # we prevent erasing a loader with None or a different loader
                            # Note one loader will aggregate root entries like repo/msg and repo/pkg/msg
                            loader = l
            # else:
            # other cases handled later by the default python pathfinder.

        return loader


class ROSDirectoryFinder(FileFinder):
    """Finder to interpret directories as modules, and files as classes"""

    def __init__(self, path, *ros_loader_details):
        """
        Finder to get directories 'msg' & 'srv' containing ROS message and service files.
        It need to be inserted in sys.path_hooks before FileFinder, since these are Directories
        but not containing __init__ as per python (hardcoded) convention.

        Note: There is a matching issue between msg/ folder and msg/My.msg on one side,
         and package, module, class concepts on the other.
        Since a module is not callable, but we need to call My(data) to build a message class (ROS convention),
         we match the msg/ folder to a module (and not a package)
        And to keep matching ROS conventions, a directory without __init__ or any message/service file,
        will become a namespace (sub)package.

        :param path_entry: the msg or srv directory path (no finder should have been instantiated yet)
        """

        ros_loaders = []
        for loader, suffixes in ros_loader_details:
            ros_loaders.extend((suffix, loader) for suffix in suffixes)
        self._ros_loaders = ros_loaders

        # We need to check that we will be able to find a module or package,
        # or raise ImportError to allow other finders to be instantiated for this path.
        # => the logic must correspond to find_module()
        findable = False
        for f in [p for p in os.listdir(path) if os.path.isdir(os.path.join(path,p))]:  # we are only interested in directories
            findable = findable or any(  # we make sure we have at least a directory that :
                    f == l.get_origin_subdir() and  # has the right name and
                    [subf for subf in os.listdir(os.path.join(path, f)) if subf.endswith(s)]
                    # contain at least one file with the right extension
                    for s, l in self._ros_loaders
            )
        # Note that testing for extensions of file in path is already too late here,
        # since we generate the whole directory at one time, and each file is a class (not a module)

        if not findable:
            raise _ImportError(
                "cannot find any matching module based on extensions {0} or origin subdirs {1} ".format(
                    [s for s, _ in self._ros_loaders], [l.get_origin_subdir() for _, l in self._ros_loaders]),
                path=path
            )
            # This is needed to not override default behavior in path where there is NO ROS files/directories.

        self.path = path or '.'

    def __repr__(self):
        return 'ROSDirectoryFinder({!r})'.format(self.path)

    @classmethod
    def path_hook(cls, *loader_details):
        def rosimporter_path_hook(path):
            """Path hook for ROSDirectoryFinder."""
            if not os.path.isdir(path):
                raise _ImportError('only directories are supported', path=path)
            return cls(path, *loader_details)

        return rosimporter_path_hook

    def find_spec(self, fullname, target=None):
        import importlib.util  # called only in python 3, which should have this

        loader = self.find_module(fullname)
        spec = importlib.util.spec_from_file_location(
            # will extract all useful information from loader
            fullname,
            loader=loader,
        )
        return spec

    def find_module(self, fullname, path=None):
        """
        Try to find a spec for the specified module.
                Returns the matching spec, or None if not found.
        :param fullname: the name of the package we are trying to import
        :param target: what we plan to do with it
        :return:
        """

        tail_module = fullname.rpartition('.')[2]
        loader = None
        base_path = os.path.join(self.path, tail_module)

        # special code here since FileFinder expect a "__init__" that we don't need for msg or srv.
        if os.path.isdir(base_path):
            loader_class = None
            rosdir = None
            # Figuring out if we should care about this directory at all
            for root, dirs, files in os.walk(base_path):
                for suffix, loader_cls in self._ros_loaders:
                    if any(f.endswith(suffix) for f in files):
                        loader_class = loader_cls
                        rosdir = root
                        break  # breaking as soon as we find something interesting
                        # we need to take care of msgs before srvs (because they can be used as dependencies
            if loader_class and rosdir and rosdir == base_path:  # we found a message/service file in the hierarchy, that belong to our module
                # rospackage = fullname.partition('.')[0]
                # # We should reproduce package structure in generated file structure
                # dirlist = base_path.split(os.sep)
                # pkgidx = dirlist[::-1].index(rospackage)
                # # find root package folder
                # pkgrootdir = os.path.join('/' if dirlist[0] == '' else '', *dirlist[:len(dirlist) - pkgidx])

                # # Find the the public ROS interface definition (that we can also depend on)
                # if not os.path.exists(os.path.join(pkgrootdir, 'package.xml')):
                #     public_msg = None
                #     # we try to find a lower 'msg' to use as dependency (only if we are not at the 'package' level, or at root (dirname(d) == d) )
                #     while os.path.dirname(pkgrootdir) != pkgrootdir and not os.path.exists(
                #             os.path.join(pkgrootdir, 'package.xml')):
                #         parentdirmsg = os.path.join(os.path.dirname(pkgrootdir), 'msg')
                #         if os.path.exists(parentdirmsg) and [f for f in os.listdir(parentdirmsg) if f.endswith('.msg')]:
                #             public_msg = parentdirmsg
                #         pkgrootdir = os.path.dirname(pkgrootdir)
                #     # TODO : maybe discover and load all messages at a "closer to root" level in package tree ??
                #     if public_msg:
                #         # we need to load the public interface first (to avoid caching an ungenerated root module)
                #         public_msg_spec = self.build_spec_loader(fullname, base_path, loader_class)


                # we are looking for submodules either in generated location
                # to be able to load generated python files) or in original msg location
                loader = loader_class(fullname, base_path)
                # We DO NOT WANT TO add the generated dir in sys.path to use a python loader
                # since the plan is to eventually not have to rely on files at all TODO

        # we return None if we couldn't build a loader before
        return loader


MSG_SUFFIXES = ['.msg']
SRV_SUFFIXES = ['.srv']


def get_supported_ros_loaders():
    """Returns a list of file-based module loaders.
    Each item is a tuple (loader, suffixes).
    """
    msg = ROSMsgLoader, MSG_SUFFIXES
    srv = ROSSrvLoader, SRV_SUFFIXES
    return [msg, srv]


def _install():
    """Install the path-based import components."""
    supported_loaders = get_supported_ros_loaders()
    sys.path_hooks.extend([ROSDirectoryFinder.path_hook(*supported_loaders)])
    # TODO : sys.meta_path.append(DistroFinder)




# Useless ?
#_ros_finder_instance_obsolete_python = ROSImportFinder

ros_distro_finder = None

# TODO : metafinder
def activate(rosdistro_path=None, *workspaces):
    global ros_distro_finder
    if rosdistro_path is None:  # autodetect most recent installed distro
        if os.path.exists('/opt/ros/lunar'):
            rosdistro_path = '/opt/ros/lunar'
        elif os.path.exists('/opt/ros/kinetic'):
            rosdistro_path = '/opt/ros/kinetic'
        elif os.path.exists('/opt/ros/jade'):
            rosdistro_path = '/opt/ros/jade'
        elif os.path.exists('/opt/ros/indigo'):
            rosdistro_path = '/opt/ros/indigo'
        else:
            raise ImportError(
                "No ROS distro detected on this system. Please specify the path when calling ROSImportMetaFinder()")

    ros_distro_finder = ROSImportMetaFinder(rosdistro_path, *workspaces)
    sys.meta_path.append(ros_distro_finder)

    #if sys.version_info >= (2, 7, 12):  # TODO : which exact version matters ?

    # We need to be before FileFinder to be able to find our (non .py[c]) files
    # inside, maybe already imported, python packages...
    sys.path_hooks.insert(1, ROSImportFinder)

    # else:  # older (trusty) version
    #     sys.path_hooks.append(_ros_finder_instance_obsolete_python)

    for hook in sys.path_hooks:
        print('Path hook: {}'.format(hook))

    # TODO : mix that with ROS PYTHONPATH shenanigans... to enable the finder only for 'ROS aware' paths
    if paths:
        sys.path.append(*paths)


def deactivate(*paths):
    """ CAREFUL : even if we remove our path_hooks, the created finder are still cached in sys.path_importer_cache."""
    #if sys.version_info >= (2, 7, 12):  # TODO : which exact version matters ?
    sys.path_hooks.remove(ROSImportFinder)
    # else:  # older (trusty) version
    #     sys.path_hooks.remove(_ros_finder_instance_obsolete_python)
    if paths:
        sys.path.remove(*paths)

    sys.meta_path.remove(ros_distro_finder)


@contextlib.contextmanager
def ROSImportContext(*paths):
    activate(*paths)
    yield
    deactivate(*paths)


# TODO : a meta finder could find a full ROS distro...
